from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
from toolz import merge, curry, assoc
from sklearn.linear_model import LinearRegression
from sklearn.gaussian_process import GaussianProcessRegressor, kernels
from sklearn import __version__ as sk_version

from fklearn.common_docstrings import learner_pred_fn_docstring, learner_return_docstring
from fklearn.types import LearnerReturnType
from fklearn.training.utils import log_learner_time, expand_features_encoded


@curry
@log_learner_time(learner_name='linear_regression_learner')
def linear_regression_learner(df: pd.DataFrame,
                              features: List[str],
                              target: str,
                              params: Dict[str, Any] = None,
                              prediction_column: str = "prediction",
                              weight_column: str = None,
                              encode_extra_cols: bool = True) -> LearnerReturnType:
    """
    Fits an linear regression classifier to the dataset. Return the predict function
    for the model and the predictions for the input dataset.

    Parameters
    ----------

    df : pandas.DataFrame
        A Pandas' DataFrame with features and target columns.
        The model will be trained to predict the target column
        from the features.

    features : list of str
        A list os column names that are used as features for the model. All this names
        should be in `df`.

    target : str
        The name of the column in `df` that should be used as target for the model.
        This column should be continuous, since this is a regression model.

    params : dict
        The LinearRegression parameters in the format {"par_name": param}. See:
        http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html

    prediction_column : str
        The name of the column with the predictions from the model.

    weight_column : str, optional
        The name of the column with scores to weight the data.

    encode_extra_cols : bool (default: True)
        If True, treats all columns in `df` with name pattern fklearn_feat__col==val` as feature columns.
    """

    def_params = {"fit_intercept": True}
    params = def_params if not params else merge(def_params, params)

    weights = df[weight_column].values if weight_column else None

    features = features if not encode_extra_cols else expand_features_encoded(df, features)

    regr = LinearRegression(**params)
    regr.fit(df[features].values, df[target].values, sample_weight=weights)

    def p(new_df: pd.DataFrame) -> pd.DataFrame:
        return new_df.assign(**{prediction_column: regr.predict(new_df[features].values)})

    p.__doc__ = learner_pred_fn_docstring("linear_regression_learner")

    log = {'linear_regression_learner': {
        'features': features,
        'target': target,
        'parameters': params,
        'prediction_column': prediction_column,
        'package': "sklearn",
        'package_version': sk_version,
        'feature_importance': dict(zip(features, regr.coef_.flatten())),
        'training_samples': len(df)},
        'object': regr}

    return p, p(df), log


linear_regression_learner.__doc__ += learner_return_docstring("Linear Regression")


@curry
@log_learner_time(learner_name='xgb_regression_learner')
def xgb_regression_learner(df: pd.DataFrame,
                           features: List[str],
                           target: str,
                           learning_rate: float = 0.1,
                           num_estimators: int = 100,
                           extra_params: Dict[str, Any] = None,
                           prediction_column: str = "prediction",
                           weight_column: str = None,
                           encode_extra_cols: bool = True) -> LearnerReturnType:
    """
    Fits an XGBoost regressor to the dataset. It first generates a DMatrix
    with the specified features and labels from `df`. Then it fits a XGBoost
    model to this DMatrix. Return the predict function for the model and the
    predictions for the input dataset.

    Parameters
    ----------

    df : pandas.DataFrame
        A Pandas' DataFrame with features and target columns.
        The model will be trained to predict the target column
        from the features.

    features : list of str
        A list os column names that are used as features for the model. All this names
        should be in `df`.

    target : str
        The name of the column in `df` that should be used as target for the model.
        This column should be numerical and continuous, since this is a regression model.

    learning_rate : float
        Float in range [0,1].
        Step size shrinkage used in update to prevents overfitting. After each boosting step,
        we can directly get the weights of new features. and eta actually shrinks the
        feature weights to make the boosting process more conservative.
        See the eta hyper-parameter in:
        http://xgboost.readthedocs.io/en/latest/parameter.html

    num_estimators : int
        Int in range [0, inf]
        Number of boosted trees to fit.
        See the n_estimators hyper-parameter in:
        http://xgboost.readthedocs.io/en/latest/python/python_api.html

    extra_params : dict, optional
        Dictionary in the format {"hyperparameter_name" : hyperparameter_value.
        Other parameters for the XGBoost model. See the list in:
        http://xgboost.readthedocs.io/en/latest/parameter.html
        If not passed, the default will be used.

    prediction_column : str
        The name of the column with the predictions from the model.

    weight_column : str, optional
        The name of the column with scores to weight the data.

    encode_extra_cols : bool (default: True)
        If True, treats all columns in `df` with name pattern fklearn_feat__col==val` as feature columns.
    """

    import xgboost as xgb

    weights = df[weight_column].values if weight_column else None
    params = extra_params if extra_params else {}
    params = assoc(params, "eta", learning_rate)
    params = params if "objective" in params else assoc(params, "objective", 'reg:linear')

    features = features if not encode_extra_cols else expand_features_encoded(df, features)

    dtrain = xgb.DMatrix(df[features].values, label=df[target].values, weight=weights, feature_names=map(str, features))

    bst = xgb.train(params, dtrain, num_estimators)

    def p(new_df: pd.DataFrame, apply_shap: bool = False) -> pd.DataFrame:
        dtest = xgb.DMatrix(new_df[features].values, feature_names=map(str, features))
        col_dict = {prediction_column: bst.predict(dtest)}

        if apply_shap:
            import shap
            explainer = shap.TreeExplainer(bst)
            shap_values = list(explainer.shap_values(new_df[features]))
            shap_expected_value = explainer.expected_value

            shap_output = {"shap_values": shap_values,
                           "shap_expected_value": np.repeat(shap_expected_value, len(shap_values))}

            col_dict = merge(col_dict, shap_output)

        return new_df.assign(**col_dict)

    p.__doc__ = learner_pred_fn_docstring("xgb_regression_learner", shap=True)

    log = {'xgb_regression_learner': {
        'features': features,
        'target': target,
        'prediction_column': prediction_column,
        'package': "xgboost",
        'package_version': xgb.__version__,
        'parameters': assoc(params, "num_estimators", num_estimators),
        'feature_importance': bst.get_score(),
        'training_samples': len(df)},
        'object': bst}

    return p, p(df), log


xgb_regression_learner.__doc__ += learner_return_docstring("XGboost Regressor")


@curry
@log_learner_time(learner_name='catboost_regressor_learner')
def catboost_regressor_learner(df: pd.DataFrame,
                               features: List[str],
                               target: str,
                               learning_rate: float = 0.1,
                               num_estimators: int = 100,
                               extra_params: Dict[str, Any] = None,
                               prediction_column: str = "prediction",
                               weight_column: str = None) -> LearnerReturnType:
    """
    Fits an CatBoost regressor to the dataset. It first generates a Pool
    with the specified features and labels from `df`. Then it fits a CatBoost
    model to this Pool. Return the predict function for the model and the
    predictions for the input dataset.

    Parameters
    ----------

    df : pandas.DataFrame
        A Pandas' DataFrame with features and target columns.
        The model will be trained to predict the target column
        from the features.

    features : list of str
        A list os column names that are used as features for the model. All this names
        should be in `df`.

    target : str
        The name of the column in `df` that should be used as target for the model.
        This column should be numerical and continuous, since this is a regression model.

    learning_rate : float
        Float in range [0,1].
        Step size shrinkage used in update to prevents overfitting. After each boosting step,
        we can directly get the weights of new features. and eta actually shrinks the
        feature weights to make the boosting process more conservative.
        See the eta hyper-parameter in:
        https://catboost.ai/docs/concepts/python-reference_parameters-list.html

    num_estimators : int
        Int in range [0, inf]
        Number of boosted trees to fit.
        See the n_estimators hyper-parameter in:
        https://catboost.ai/docs/concepts/python-reference_parameters-list.html

    extra_params : dict, optional
        Dictionary in the format {"hyperparameter_name" : hyperparameter_value.
        Other parameters for the CatBoost model. See the list in:
        https://catboost.ai/docs/concepts/python-reference_catboostregressor.html
        If not passed, the default will be used.

    prediction_column : str
        The name of the column with the predictions from the model.

    weight_column : str, optional
        The name of the column with scores to weight the data.
    """
    from catboost import Pool, CatBoostRegressor
    import catboost

    weights = df[weight_column].values if weight_column else None
    params = extra_params if extra_params else {}
    params = assoc(params, "eta", learning_rate)

    dtrain = Pool(df[features].values, df[target].values, weight=weights, feature_names=list(map(str, features)))
    cat_boost_regressor = CatBoostRegressor(iterations=num_estimators, **params)
    cbr = cat_boost_regressor.fit(dtrain, verbose=0)

    def p(new_df: pd.DataFrame, apply_shap: bool = False) -> pd.DataFrame:
        dtest = Pool(new_df[features].values, feature_names=list(map(str, features)))
        col_dict = {prediction_column: cbr.predict(dtest)}

        if apply_shap:
            import shap
            explainer = shap.TreeExplainer(cbr)
            shap_values = list(explainer.shap_values(new_df[features]))
            shap_expected_value = explainer.expected_value

            shap_output = {"shap_values": shap_values,
                           "shap_expected_value": np.repeat(shap_expected_value, len(shap_values))}

            col_dict = merge(col_dict, shap_output)

        return new_df.assign(**col_dict)

    p.__doc__ = learner_pred_fn_docstring("CatBoostRegressor", shap=False)

    log = {'catboost_regression_learner': {
        'features': features,
        'target': target,
        'prediction_column': prediction_column,
        'package': "catboost",
        'package_version': catboost.__version__,
        'parameters': assoc(params, "num_estimators", num_estimators),
        'feature_importance': cbr.feature_importances_,
        'training_samples': len(df)},
        'object': cbr}

    return p, p(df), log


catboost_regressor_learner.__doc__ += learner_return_docstring("CatBoostRegressor")


@curry
@log_learner_time(learner_name='gp_regression_learner')
def gp_regression_learner(df: pd.DataFrame,
                          features: List[str],
                          target: str,
                          kernel: kernels.Kernel = None,
                          alpha: float = 0.1,
                          extra_variance: Union[str, float] = "fit",
                          return_std: bool = False,
                          extra_params: Dict[str, Any] = None,
                          prediction_column: str = "prediction",
                          encode_extra_cols: bool = True) -> LearnerReturnType:
    """
    Fits an gaussian process regressor to the dataset.

    Parameters
    ----------

    df: pandas.DataFrame
        A Pandas' DataFrame with features and target columns.
        The model will be trained to predict the target column
        from the features.

    features: list of str
        A list os column names that are used as features for the model. All this names
        should be in `df`.

    target: str
        The name of the column in `df` that should be used as target for the model.
        This column should be numerical and continuous, since this is a regression model.

    kernel: sklearn.gaussian_process.kernels
        The kernel specifying the covariance function of the GP. If None is passed,
        the kernel "1.0 * RBF(1.0)" is used as default. Note that the kernel's hyperparameters
        are optimized during fitting.

    alpha: float
        Value added to the diagonal of the kernel matrix during fitting. Larger values correspond to increased
        noise level in the observations. This can also prevent a potential numerical issue during fitting,
        by ensuring that the calculated values form a positive definite matrix.

    extra_variance: float
        The amount of extra variance to scale to the predictions in standard deviations. If left as the default "fit",
        Uses the standard deviation of the target.

    return_std: bool
        If True, the standard-deviation of the predictive distribution at the query points is returned
        along with the mean.

    extra_params: dict {"hyperparameter_name" : hyperparameter_value}, optional
        Other parameters for the GaussianProcessRegressor model. See the list in:
        http://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.GaussianProcessRegressor.html
        If not passed, the default will be used.

    prediction_column : str
        The name of the column with the predictions from the model.

    encode_extra_cols : bool (default: True)
        If True, treats all columns in `df` with name pattern fklearn_feat__col==val` as feature columns.
    """

    params = extra_params if extra_params else {}

    params['alpha'] = alpha
    params['kernel'] = kernel

    features = features if not encode_extra_cols else expand_features_encoded(df, features)

    gp = GaussianProcessRegressor(**params)
    gp.fit(df[features], df[target])

    extra_variance = df[target].std() if extra_variance == "fit" else extra_variance if extra_variance else 1

    def p(new_df: pd.DataFrame) -> pd.DataFrame:
        if return_std:
            pred_mean, pred_std = gp.predict(df[features], return_std=True)
            pred_std *= extra_variance
            return new_df.assign(**{prediction_column: pred_mean, prediction_column + "_std": pred_std})
        else:
            return new_df.assign(**{prediction_column: gp.predict(df[features])})

    p.__doc__ = learner_pred_fn_docstring("gp_regression_learner")

    log = {'gp_regression_learner': {
        'features': features,
        'target': target,
        'parameters': merge(params, {'extra_variance': extra_variance,
                                     'return_std': return_std}),
        'prediction_column': prediction_column,
        'package': "sklearn",
        'package_version': sk_version,
        'training_samples': len(df)},
        'object': gp}

    return p, p(df), log


gp_regression_learner.__doc__ += learner_return_docstring("Gaussian Process Regressor")


@curry
@log_learner_time(learner_name='lgbm_regression_learner')
def lgbm_regression_learner(df: pd.DataFrame,
                            features: List[str],
                            target: str,
                            learning_rate: float = 0.1,
                            num_estimators: int = 100,
                            extra_params: Dict[str, Any] = None,
                            prediction_column: str = "prediction",
                            weight_column: str = None,
                            encode_extra_cols: bool = True) -> LearnerReturnType:
    """
    Fits an LGBM regressor to the dataset.

    It first generates a Dataset with the specified features and labels
    from `df`. Then, it fits a LGBM model to this Dataset. Return the predict
    function for the model and the predictions for the input dataset.

    Parameters
    ----------

    df : pandas.DataFrame
        A Pandas' DataFrame with features and target columns.
        The model will be trained to predict the target column
        from the features.

    features : list of str
        A list os column names that are used as features for the model. All this names
        should be in `df`.

    target : str
        The name of the column in `df` that should be used as target for the model.
        This column should be binary, since this is a classification model.

    learning_rate : float
        Float in the range (0, 1]
        Step size shrinkage used in update to prevents overfitting. After each boosting step,
        we can directly get the weights of new features. and eta actually shrinks the
        feature weights to make the boosting process more conservative.
        See the learning_rate hyper-parameter in:
        https://github.com/Microsoft/LightGBM/blob/master/docs/Parameters.rst

    num_estimators : int
        Int in the range (0, inf)
        Number of boosted trees to fit.
        See the num_iterations hyper-parameter in:
        https://github.com/Microsoft/LightGBM/blob/master/docs/Parameters.rst

    extra_params : dict, optional
        Dictionary in the format {"hyperparameter_name" : hyperparameter_value}.
        Other parameters for the LGBM model. See the list in:
        https://github.com/Microsoft/LightGBM/blob/master/docs/Parameters.rst
        If not passed, the default will be used.

    prediction_column : str
        The name of the column with the predictions from the model.

    weight_column : str, optional
        The name of the column with scores to weight the data.

    encode_extra_cols : bool (default: True)
        If True, treats all columns in `df` with name pattern fklearn_feat__col==val` as feature columns.
    """

    import lightgbm as lgbm

    params = extra_params if extra_params else {}
    params = assoc(params, "eta", learning_rate)
    params = params if "objective" in params else assoc(params, "objective", 'regression')

    weights = df[weight_column].values if weight_column else None

    features = features if not encode_extra_cols else expand_features_encoded(df, features)

    dtrain = lgbm.Dataset(df[features].values, label=df[target], feature_name=list(map(str, features)), weight=weights,
                          silent=True)

    bst = lgbm.train(params, dtrain, num_estimators)

    def p(new_df: pd.DataFrame, apply_shap: bool = False) -> pd.DataFrame:
        col_dict = {prediction_column: bst.predict(new_df[features].values)}

        if apply_shap:
            import shap
            explainer = shap.TreeExplainer(bst)
            shap_values = list(explainer.shap_values(new_df[features]))
            shap_expected_value = explainer.expected_value

            shap_output = {"shap_values": shap_values,
                           "shap_expected_value": np.repeat(shap_expected_value, len(shap_values))}

            col_dict = merge(col_dict, shap_output)

        return new_df.assign(**col_dict)

    p.__doc__ = learner_pred_fn_docstring("lgbm_regression_learner", shap=True)

    log = {'lgbm_regression_learner': {
        'features': features,
        'target': target,
        'prediction_column': prediction_column,
        'package': "lightgbm",
        'package_version': lgbm.__version__,
        'parameters': assoc(params, "num_estimators", num_estimators),
        'feature_importance': dict(zip(features, bst.feature_importance().tolist())),
        'training_samples': len(df)},
        'object': bst}

    return p, p(df), log


lgbm_regression_learner.__doc__ += learner_return_docstring("LGBM Regressor")


@curry
@log_learner_time(learner_name='custom_supervised_model_learner')
def custom_supervised_model_learner(df: pd.DataFrame,
                                    features: List[str],
                                    target: str,
                                    model: Any,
                                    supervised_type: str,
                                    log: Dict[str, Dict],
                                    prediction_column: str = "prediction") -> LearnerReturnType:
    """
    Fits a custom model to the dataset.
    Return the predict function, the predictions for the input dataset and a log describing the model.

    Parameters
    ----------

    df : pandas.DataFrame
        A Pandas' DataFrame with features and target columns.
        The model will be trained to predict the target column
        from features.

    features : list of str
        A list os column names that are used as features for the model. All this names
        should be in `df`.

    target : str
        The name of the column in `df` that should be used as target for the model.

    model: Object
        Machine learning model to be used for regression or clasisfication.
        model object must have ".fit" attribute to train the data.
        For classification problems, it also needs ".predict_proba" attribute.
        For regression problemsm it needs ".predict" attribute.

    supervised_type: str
        Type of supervised learning to be used
        The options are: 'classification' or 'regression'

    log : Dict[str, Dict]
        Log with additional information of the custom model used.
        It must start with just one element with the model name.

    prediction_column : str
        The name of the column with the predictions from the model.
        For classification problems, all probabilities wiill be added: for i in range(0,n_classes).
        For regression just prediction_column will be added.
    """

    if len(log) != 1:
        raise ValueError("\'log\' dictionary must start with model name")
    if supervised_type not in ('classification', 'regression'):
        raise TypeError("supervised_type options are: \'classification\' or \'regression\'")
    if not hasattr(model, 'fit'):
        raise AttributeError("\'model\' object must have \'fit\' attribute")
    if supervised_type == 'classification' and not hasattr(model, 'predict_proba'):
        raise AttributeError("\'model\' object for classification must have \'predict_proba\' attribute")
    if supervised_type == 'regression' and not hasattr(model, 'predict'):
        raise AttributeError("\'model\' object for regression must have \'predict\' attribute")

    model.fit(df[features].values, df[target].values)

    def p(new_df: pd.DataFrame) -> pd.DataFrame:
        if supervised_type == 'classification':
            pred = model.predict_proba(new_df[features].values)
            col_dict = {}
            for (key, value) in enumerate(pred.T):
                col_dict.update({prediction_column + "_" + str(key): value})
        elif supervised_type == 'regression':
            col_dict = {prediction_column: model.predict(new_df[features].values)}

        return new_df.assign(**col_dict)

    p.__doc__ = learner_pred_fn_docstring("custom_supervised_model_learner")

    log["object"] = model

    return p, p(df), log


custom_supervised_model_learner.__doc__ += learner_return_docstring("Custom Supervised Model Learner")



from typing import Any, Dict, Generator, List

from toolz.curried import reduce, partial, pipe, first, curry

from fklearn.metrics.pd_extractors import extract
from fklearn.types import LogListType, LogType, ExtractorFnType, ValidatorReturnType, EvalReturnType


@curry
def get_avg_metric_from_extractor(logs: LogType, extractor: ExtractorFnType, metric_name: str) -> float:
    metric_folds = extract(logs["validator_log"], extractor)
    return metric_folds[metric_name].mean()


def get_best_performing_log(log_list: LogListType, extractor: ExtractorFnType, metric_name: str) -> Dict:
    logs_eval = [get_avg_metric_from_extractor(log, extractor, metric_name) for log in log_list]
    return pipe(logs_eval, partial(zip, log_list), partial(sorted, reverse=True, key=lambda x: x[1]))[0][0]


def get_used_features(log: Dict) -> List[str]:
    return first((gen_dict_extract('features', log)))


def order_feature_importance_avg_from_logs(log: Dict) -> List[str]:
    d = first(gen_dict_extract('feature_importance', log))
    return sorted(d, key=d.get, reverse=True)


def gen_key_avgs_from_logs(key: str, logs: List[Dict]) -> Dict[str, float]:
    return gen_key_avgs_from_dicts([gen_key_avgs_from_iteration(key, log) for log in logs])


def gen_key_avgs_from_iteration(key: str, log: Dict) -> Any:
    return first(gen_dict_extract(key, log))


def gen_key_avgs_from_dicts(obj: List) -> Dict[str, float]:
    sum_values_by_key = reduce(lambda x, y: dict((k, v + y.get(k, 0)) for k, v in x.items()), obj)
    return {k: float(v) / len(obj) for k, v in sum_values_by_key.items()}


def gen_dict_extract(key: str, obj: Dict) -> Generator[Any, None, None]:
    if hasattr(obj, 'items'):
        for k, v in obj.items():
            if k == key:
                yield v
            if isinstance(v, dict):
                for result in gen_dict_extract(key, v):
                    yield result
            elif isinstance(v, list):
                for d in v:
                    for result in gen_dict_extract(key, d):
                        yield result


@curry
def gen_validator_log(eval_log: EvalReturnType, fold_num: int, test_size: int) -> ValidatorReturnType:
    return {'validator_log': [{'fold_num': fold_num, 'split_log': {'test_size': test_size},
                               'eval_results': [eval_log]}]}



import random
from typing import List

import numpy as np
from toolz import curry
import pandas as pd

from fklearn.types import ColumnWisePerturbFnType


@curry
def shift_mu(col: pd.Series, perc: float) -> pd.Series:
    """
    Shift the mean of column by a given percentage

    Parameters
    ----------
    col : pd.Series
        A Pandas' Series

    perc : float
        How much to shift the mu percentually (can be negative)

    Returns
    ----------
    A transformed pd.Series
    """
    mu = np.mean(col)
    col = col + mu * perc
    return col


@curry
def random_noise(col: pd.Series, mag: float) -> pd.Series:
    """
    Fit a gaussian to column, then sample and add to each entry with a magnification parameter

    Parameters
    ----------
    col : pd.Series
        A Pandas' Series

    mag : float
        Multiplies the noise to control scaling

    Returns
    ----------
    A transformed pd.Series
    """
    mu = np.mean(col)
    std = np.std(col)
    noise = np.random.normal(mu, std, len(col)) * mag
    return col + noise


@curry
def nullify(col: pd.Series, perc: float = 1) -> pd.Series:
    """
    Replace a percenteage of values in  the input Series by np.nan

    Parameters
    ----------
    col : pd.Series
        A Pandas' Series

    perc : float
        Percentage to be replaced by no.nan

    Returns
    ----------
    A transformed pd.Series
    """
    # default behavior to nullify whole column
    n = len(col)
    ix_to_nan = random.sample(range(n), int(n * perc))
    ret = col.copy(deep=True)
    ret.iloc[ix_to_nan] = np.nan
    return ret


@curry
def sample_columns(data: pd.DataFrame, perc: float) -> List[str]:
    """
    Helper function that picks randomly a percentage of the columns

    Parameters
    ----------
    data : pd.DataFrame
        A Pandas' DataFrame

    perc : float
        Percentage of columns to be sampled

    Returns
    ----------
    A list of column names
    """
    return random.sample(list(data.columns), int(len(data.columns) * perc))


@curry
def perturbator(data: pd.DataFrame,
                cols: List[str],
                corruption_fn: ColumnWisePerturbFnType) -> pd.DataFrame:
    """
    transforms specific columns of a dataset according to an artificial
    corruption function.

    Parameters
    ----------
    data : pandas.DataFrame
        A Pandas' DataFrame

    cols : List[str]
        A list of columns to apply the corruption function

    corruption_fn : function pandas.Series -> pandas.Series
        An arbitrary corruption function

    Returns
    ----------
    A transformed dataset
    """
    return data.assign(**{col: corruption_fn(data[col]) for col in cols})



from typing import Callable, Dict, List

from toolz.curried import pipe, first, mapcat
import pandas as pd

from fklearn.tuning.samplers import remove_features_subsets, remove_by_feature_importance, remove_by_feature_shuffling
from fklearn.tuning.stoppers import stop_by_num_features, stop_by_num_features_parallel, stop_by_iter_num, \
    stop_by_no_improvement, stop_by_no_improvement_parallel, aggregate_stop_funcs
from fklearn.validation.validator import parallel_validator
from fklearn.types import EvalFnType, ExtractorFnType, LearnerReturnType, ListLogListType, LogListType, SplitterFnType,\
    ValidatorReturnType, LogType

SaveIntermediaryFnType = Callable[[List[ValidatorReturnType]], None]
TuningLearnerFnType = Callable[[pd.DataFrame, List[str]], LearnerReturnType]


def feature_importance_backward_selection(train_data: pd.DataFrame,
                                          param_train_fn: TuningLearnerFnType,
                                          features: List[str],
                                          split_fn: SplitterFnType,
                                          eval_fn: EvalFnType,
                                          extractor: ExtractorFnType,
                                          metric_name: str,
                                          num_removed_by_step: int = 5,
                                          threshold: float = 0.005,
                                          early_stop: int = 2,
                                          iter_limit: int = 50,
                                          min_remaining_features: int = 50,
                                          save_intermediary_fn: SaveIntermediaryFnType = None,
                                          n_jobs: int = 1) -> ListLogListType:
    """
        Performs train-evaluation iterations while subsampling the used features
        to compute statistics about feature relevance

        Parameters
        ----------
        train_data : pandas.DataFrame
            A Pandas' DataFrame with training data

        auxiliary_columns: list of str
            List of columns from the dataset that are not used as features but are
            used for evaluation or cross validation. (id, date, etc)

        param_train_fn : function (DataFrame, List of Strings) -> prediction_function, predictions_dataset, logs
            A partially defined learning function that takes a training set and a feature list and
            returns a predict function, a dataset with training predictions and training
            logs.

        features: list of str
            Elements must be columns of the train_data

        split_fn : function pandas.DataFrame ->  list of tuple
            Partially defined split function that takes a dataset and returns
            a list of folds. Each fold is a Tuple of arrays. The fist array in
            each tuple contains training indexes while the second array
            contains validation indexes.

        eval_fn : function pandas.DataFrame -> dict
            A partially defined evaluation function that takes a dataset with prediction and
            returns the evaluation logs.

        extractor: function str -> float
            A extractor that take a string and returns the value of that string on a dict

        metric_name: str
            String with the name of the column that refers to the metric column to be extracted

        num_removed_by_step: int (default 5)
            Number of features removed at each iteration

        threshold: float (default 0.005)
            Threshold for model performance comparison

        early_stop: int (default 2)
            Number of rounds without improvement before stopping process

        iter_limit: int (default 50)
            Maximum number of iterations before stopping

        min_remaining_features: int (default 50)
            Minimum number of features that should remain in the model,
            combining num_removed_by_step and iter_limit accomplishes the same
            functionality as this parameter.

        save_intermediary_fn : function(log) -> save to file
            Partially defined saver function that receives a log result from a
            tuning step and appends it into a file
            Example: save_intermediary_result(save_path='tuning.pkl')

        n_jobs : int
            Number of parallel processes to spawn.

        Returns
        ----------
        Logs: list of list of dict
            A list log-like lists of dictionaries evaluations. Each element of the
            list is validation step of the algorithm.

    """

    selector_fn = remove_by_feature_importance(num_removed_by_step=num_removed_by_step)

    stop_fn = aggregate_stop_funcs(
        stop_by_no_improvement(extractor=extractor, metric_name=metric_name, early_stop=early_stop,
                               threshold=threshold),
        stop_by_iter_num(iter_limit=iter_limit),
        stop_by_num_features(min_num_features=min_remaining_features))

    train_fn = lambda df: param_train_fn(df, features)
    first_logs = parallel_validator(train_data, split_fn, train_fn, eval_fn, n_jobs=n_jobs)

    logs = [first_logs]
    while not stop_fn(logs):
        curr_log = first(logs)

        new_features = selector_fn(curr_log)
        new_train_fn = lambda df: param_train_fn(df, new_features)
        next_log = parallel_validator(train_data, split_fn, new_train_fn, eval_fn, n_jobs=n_jobs)

        if save_intermediary_fn is not None:
            save_intermediary_fn(next_log)

        logs = [next_log] + logs

    return logs


def poor_man_boruta_selection(train_data: pd.DataFrame,
                              test_data: pd.DataFrame,
                              param_train_fn: TuningLearnerFnType,
                              features: List[str],
                              eval_fn: EvalFnType,
                              extractor: ExtractorFnType,
                              metric_name: str,
                              max_removed_by_step: int = 5,
                              threshold: float = 0.005,
                              early_stop: int = 2,
                              iter_limit: int = 50,
                              min_remaining_features: int = 50,
                              save_intermediary_fn: Callable[[LogType], None] = None,
                              speed_up_by_importance: bool = False,
                              parallel: bool = False,
                              nthread: int = 1,
                              seed: int = 7) -> LogListType:
    """
        Performs train-evaluation iterations while shuffiling the used features
        to compute statistics about feature relevance

        Parameters
        ----------
        train_data : pandas.DataFrame
            A Pandas' DataFrame with training data

        test_data : pandas.DataFrame
            A Pandas' DataFrame with test data

        param_train_fn : function (pandas.DataFrame, list of str) -> prediction_function, predictions_dataset, logs
            A partially defined AND curried learning function that takes a training set and a feature list and
            returns a predict function, a dataset with training predictions and training
            logs.

        features: list of str
            Elements must be columns of the train_data

        eval_fn : function pandas.DataFrame -> dict
            A partially defined evaluation function that takes a dataset with prediction and
            returns the evaluation logs.

        extractor: function str -> float
            A extractor that take a string and returns the value of that string on a dict

        metric_name: str
            String with the name of the column that refers to the metric column to be extracted

        max_removed_by_step: int (default 5)
            The maximum number of features to remove. It will only consider the least max_removed_by_step in terms of
            feature importance. If speed_up_by_importance=True it will first filter the least relevant feature an
            shuffle only those. If speed_up_by_importance=False it will shuffle all features and drop the last
            max_removed_by_step in terms of PIMP. In both cases, the features will only be removed if drop in
            performance is up to the defined threshold.

        threshold: float (default 0.005)
            Threshold for model performance comparison

        early_stop: int (default 2)
            Number of rounds without improvement before stopping process

        iter_limit: int (default 50)
            Maximum number of iterations before stopping

        min_remaining_features: int (default 50)
            Minimum number of features that should remain in the model,
            combining num_removed_by_step and iter_limit accomplishes the same
            functionality as this parameter.

        save_intermediary_fn: function(log) -> save to file
            Partially defined saver function that receives a log result from a
            tuning step and appends it into a file
            Example: save_intermediary_result(save_path='tuning.pkl')

        speed_up_by_importance: bool (default True)
            If it should narrow search looking at feature importance first before getting PIMP importance. If True,
            will only shuffle the top num_removed_by_step in terms of feature importance.

        max_removed_by_step: int (default 50)
            If speed_up_by_importance=False, this will limit the number of features dropped by iteration. It will only
            drop the max_removed_by_step features that decrease the metric by the least when dropped.

        parallel: bool (default False)
            Run shuffling and prediction in parallel. Only applies if speed_up_by_importance=False

        nthread: int (default 1)
            Number of threads to run predictions. ONly applied if speed_up_by_importance=False

        seed: int (default 7)
            random state for consistency.


        Returns
        ----------
        logs: list of list of dict
            A list log-like lists of dictionaries evaluations. Each element of the
            list is validation step of the algorithm.

    """

    selector_fn = remove_by_feature_shuffling(eval_fn=eval_fn,
                                              eval_data=test_data,
                                              extractor=extractor,
                                              metric_name=metric_name,
                                              max_removed_by_step=max_removed_by_step,
                                              threshold=threshold,
                                              speed_up_by_importance=speed_up_by_importance,
                                              parallel=parallel,
                                              nthread=nthread,
                                              seed=seed)

    stop_fn = aggregate_stop_funcs(
        stop_by_no_improvement(extractor=extractor, metric_name=metric_name, early_stop=early_stop,
                               threshold=threshold),
        stop_by_iter_num(iter_limit=iter_limit),
        stop_by_num_features(min_num_features=min_remaining_features)
    )

    predict_fn_first, _, train_logs = param_train_fn(train_data, features)
    eval_logs = eval_fn(predict_fn_first(test_data))

    first_logs = {
        'train_log': train_logs,
        'validator_log': [
            {
                'fold_num': 0,
                'split_log': {
                    'train_size': train_data.shape[0],
                    'test_size': test_data.shape[0]
                },
                'eval_results': [eval_logs]
            }
        ]
    }

    logs = [first_logs]
    predict_fn = predict_fn_first

    while not stop_fn(logs):  # type: ignore
        next_features = pipe(logs, first, selector_fn(predict_fn=predict_fn))

        if len(next_features) == 0:
            break

        next_predict_fn, _, next_train_logs = param_train_fn(train_data, next_features)

        eval_logs = pipe(test_data, next_predict_fn, eval_fn)
        next_log = {'train_log': next_train_logs, 'validator_log': [
            {'fold_num': 0, 'split_log': {'train_size': train_data.shape[0], 'test_size': test_data.shape[0]},
             'eval_results': [eval_logs]}]}

        logs = [next_log] + logs

        if save_intermediary_fn is not None:
            save_intermediary_fn(next_log)

        predict_fn = next_predict_fn

    return logs


def backward_subset_feature_selection(train_data: pd.DataFrame,
                                      param_train_fn: TuningLearnerFnType,
                                      features_sets: Dict[str, List[str]],
                                      split_fn: SplitterFnType,
                                      eval_fn: EvalFnType,
                                      extractor: ExtractorFnType,
                                      metric_name: str,
                                      threshold: float = 0.005,
                                      num_removed_by_step: int = 3,
                                      early_stop: int = 2,
                                      iter_limit: int = 50,
                                      min_remaining_features: int = 50,
                                      save_intermediary_fn: SaveIntermediaryFnType = None,
                                      n_jobs: int = 1) -> ListLogListType:
    """
        Performs train-evaluation iterations while testing the subsets of features
        to compute statistics about the importance of each feature category

        Parameters
        ----------
        train_data : pandas.DataFrame
            A Pandas' DataFrame with training data

        param_train_fn : function (pandas.DataFrame, list of str) -> prediction_function, predictions_dataset, logs
            A partially defined learning function that takes a training set and a feature list and
            returns a predict function, a dataset with training predictions and training
            logs.

        features_sets: dict of string -> list
            Each String Key on the dict is a subset of columns from the dataset, the function will
            analyse the influence of each group of features on the model performance

        split_fn : function pandas.DataFrame ->  list of tuple
            Partially defined split function that takes a dataset and returns
            a list of folds. Each fold is a Tuple of arrays. The fist array in
            each tuple contains training indexes while the second array
            contains validation indexes.

        eval_fn : function pandas.DataFrame -> dict
            A partially defined evaluation function that takes a dataset with prediction and
            returns the evaluation logs.

        extractor: function str -> float
            A extractor that take a string and returns the value of that string on a dict

        metric_name: str
            String with the name of the column that refers to the metric column to be extracted

        num_removed_by_step: int (default 3)
            Number of features removed at each iteration

        threshold: float (default 0.005)
            Threshold for model performance comparison

        early_stop: int (default 2)
            Number of rounds without improvement before stopping process

        iter_limit: int (default 50)
            Maximum number of iterations before stopping

        min_remaining_features: int (default 50)
            Minimum number of features that should remain in the model,
            combining num_removed_by_step and iter_limit accomplishes the same
            functionality as this parameter.

        save_intermediary_fn : function(log) -> save to file
            Partially defined saver function that receives a log result from a
            tuning step and appends it into a file
            Example: save_intermediary_result(save_path='tuning.pkl')

        n_jobs : int
            Number of parallel processes to spawn.

        Returns
        ----------
        logs: list of list of dict
            A list log-like lists of dictionaries evaluations. Each element of the
            list is validation step of the algorithm.

    """

    selector_fn = remove_features_subsets(extractor=extractor,
                                          metric_name=metric_name,
                                          num_removed_by_step=num_removed_by_step)

    stop_fn = aggregate_stop_funcs(
        stop_by_no_improvement_parallel(extractor=extractor, metric_name=metric_name, early_stop=early_stop,
                                        threshold=threshold),
        stop_by_iter_num(iter_limit=iter_limit),
        stop_by_num_features_parallel(extractor=extractor, metric_name=metric_name,
                                      min_num_features=min_remaining_features)
    )

    used_subsets = [features_sets.keys()]

    used_features = [list(mapcat(lambda key: features_sets[key], subset)) for subset in used_subsets]

    trainers = [lambda df: param_train_fn(df, feat) for feat in used_features]

    first_val_logs = [parallel_validator(train_data, split_fn, train_func, eval_fn, n_jobs) for train_func in trainers]
    logs = [[dict(log, **{"used_subsets": list(subset)}) for log, subset in zip(first_val_logs, used_subsets)]]

    while not stop_fn(logs):
        curr_log = first(logs)

        new_subsets = selector_fn(curr_log)
        new_features = [list(mapcat(lambda key: features_sets[key], subset)) for subset in new_subsets]

        trainers = [lambda df: param_train_fn(df, feat) for feat in new_features]

        val_logs = [parallel_validator(train_data, split_fn, train_func, eval_fn, n_jobs) for train_func in trainers]

        new_logs = [dict(log, **{"used_subsets": subset}) for log, subset in zip(val_logs, new_subsets)]

        if save_intermediary_fn is not None:
            save_intermediary_fn(new_logs)

        logs = [new_logs] + logs

    return logs



import gc
from itertools import combinations
from typing import List, Tuple

import pandas as pd
from joblib import Parallel, delayed
from numpy import random
from toolz.curried import curry, first, compose, valfilter, sorted, pipe, take

from fklearn.tuning.utils import order_feature_importance_avg_from_logs, get_best_performing_log, gen_dict_extract, \
    get_avg_metric_from_extractor, get_used_features, gen_validator_log
from fklearn.types import EvalFnType, ExtractorFnType, LogListType, LogType, PredictFnType


@curry
def remove_by_feature_importance(log: LogType,
                                 num_removed_by_step: int = 5) -> List[str]:
    """
        Performs feature selection based on feature importance

        Parameters
        ----------
        log : dict
            Dictionaries evaluations.

        num_removed_by_step: int (default 5)
            The number of features to remove

        Returns
        ----------
        features: list of str
            The remaining features after removing based on feature importance

    """
    return order_feature_importance_avg_from_logs(log)[:-num_removed_by_step]


@curry
def remove_features_subsets(log_list: LogListType,
                            extractor: ExtractorFnType,
                            metric_name: str,
                            num_removed_by_step: int = 1) -> List[Tuple[str, ...]]:
    """
        Performs feature selection based on the best performing model out of
        several trained models

        Parameters
        ----------
        log_list : list of dict
            A list of log-like lists of dictionaries evaluations.

        extractor: function string -> float
            A extractor that take a string and returns the value of that string on a dict

        metric_name: str
            String with the name of the column that refers to the metric column to be extracted

        num_removed_by_step: int (default 1)
            The number of features to remove

        Returns
        ----------
        keys: list of str
            The remaining keys of feature sets after choosing the current best subset

    """

    best_log = get_best_performing_log(log_list, extractor, metric_name)
    best_subset: List[str] = first(gen_dict_extract('used_subsets', best_log))

    return list(combinations(best_subset, len(best_subset) - num_removed_by_step))


@curry
def remove_by_feature_shuffling(log: LogType,
                                predict_fn: PredictFnType,
                                eval_fn: EvalFnType,
                                eval_data: pd.DataFrame,
                                extractor: ExtractorFnType,
                                metric_name: str,
                                max_removed_by_step: int = 50,
                                threshold: float = 0.005,
                                speed_up_by_importance: bool = False,
                                parallel: bool = False,
                                nthread: int = 1,
                                seed: int = 7) -> List[str]:

    """
        Performs feature selection based on the evaluation of the test vs the
        evaluation of the test with randomly shuffled features

        Parameters
        ----------
        log : LogType
            Dictionaries evaluations.

        predict_fn: function pandas.DataFrame -> pandas.DataFrame
            A partially defined predictor that takes a DataFrame and returns the
            predicted score for this dataframe

        eval_fn : function DataFrame -> log dict
            A partially defined evaluation function that takes a dataset with prediction and
            returns the evaluation logs.

        eval_data: pandas.DataFrame
            Data used to evaluate the model after shuffling

        extractor: function str -> float
            A extractor that take a string and returns the value of that string on a dict

        metric_name: str
            String with the name of the column that refers to the metric column to be extracted

        max_removed_by_step: int (default 5)
            The maximum number of features to remove. It will only consider the least max_removed_by_step in terms of
            feature importance. If speed_up_by_importance=True it will first filter the least relevant feature an
            shuffle only those. If speed_up_by_importance=False it will shuffle all features and drop the last
            max_removed_by_step in terms of PIMP. In both cases, the features will only be removed if drop in
            performance is up to the defined threshold.

        threshold: float (default 0.005)
            Threshold for model performance comparison

        speed_up_by_importance: bool (default True)
            If it should narrow search looking at feature importance first before getting PIMP importance. If True,
            will only shuffle the top num_removed_by_step in terms of feature importance.

        parallel: bool (default False)

        nthread: int (default 1)

        seed: int (default 7)
            Random seed

        Returns
        ----------
        features: list of str
            The remaining features after removing based on feature importance

    """
    random.seed(seed)

    curr_metric = get_avg_metric_from_extractor(log, extractor, metric_name)
    eval_size = eval_data.shape[0]

    features_to_shuffle = order_feature_importance_avg_from_logs(log)[-max_removed_by_step:] \
        if speed_up_by_importance else get_used_features(log)

    def shuffle(feature: str) -> pd.DataFrame:
        return eval_data.assign(**{feature: eval_data[feature].sample(frac=1.0)})

    feature_to_delta_metric = compose(lambda m: curr_metric - m,
                                      get_avg_metric_from_extractor(extractor=extractor, metric_name=metric_name),
                                      gen_validator_log(fold_num=0, test_size=eval_size), eval_fn, predict_fn, shuffle)

    if parallel:
        metrics = Parallel(n_jobs=nthread, backend="threading")(
            delayed(feature_to_delta_metric)(feature) for feature in features_to_shuffle)
        feature_to_delta_metric = dict(zip(features_to_shuffle, metrics))
        gc.collect()

    else:
        feature_to_delta_metric = {feature: feature_to_delta_metric(feature) for feature in features_to_shuffle}

    return pipe(feature_to_delta_metric,
                valfilter(lambda delta_metric: delta_metric < threshold),
                sorted(key=lambda f: feature_to_delta_metric.get(f)),
                take(max_removed_by_step),
                list)



from os.path import dirname, join


def version() -> str:
    """
    Get package version

    Returns
    -------
    version : str
    """
    with open(join(dirname(__file__), 'resources', 'VERSION')) as f:
        return f.read().strip()


__version__ = version()



from typing import Optional, Tuple

import numpy as np
from numpy.random import RandomState
import pandas as pd
from sklearn.model_selection import StratifiedShuffleSplit
from toolz import curry

from fklearn.types import DateType


@curry
def time_split_dataset(dataset: pd.DataFrame,
                       train_start_date: DateType,
                       train_end_date: DateType,
                       holdout_end_date: DateType,
                       time_column: str,
                       holdout_start_date: DateType = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Splits temporal data into a training and testing datasets such that
    all training data comes before the testings one.

    Parameters
    ----------
    dataset : pandas.DataFrame
        A Pandas' DataFrame with an Identifier Column and a Date Column.
        The model will be trained to predict the target column
        from the features.

    train_start_date : str
        A date string representing a the starting time of the training data.
        It should be in the same format as the Date Column in `dataset`.

    train_end_date : str
        A date string representing a the ending time of the training data.
        This will also be used as the start date of the holdout period if no `holdout_start_date` is given.
        It should be in the same format as the Date Column in `dataset`.

    holdout_end_date : str
        A date string representing a the ending time of the holdout data.
        It should be in the same format as the Date Column in `dataset`.

    time_column : str
        The name of the Date column of `dataset`.

    holdout_start_date: str
        A date string representing the starting time of the holdout data.
        If `None` is given it will be equal to `train_end_date`.
        It should be in the same format as the Date Column in `dataset`.

    Returns
    ----------
    train_set : pandas.DataFrame
        The in ID sample and in time training set.

    test_set : pandas.DataFrame
        The out of ID sample and in time hold out set.
    """

    holdout_start_date = holdout_start_date if holdout_start_date else train_end_date

    train_set = dataset[
        (dataset[time_column] >= train_start_date) & (dataset[time_column] < train_end_date)]

    test_set = dataset[
        (dataset[time_column] >= holdout_start_date) & (dataset[time_column] < holdout_end_date)]

    return train_set, test_set


@curry
def space_time_split_dataset(dataset: pd.DataFrame,
                             train_start_date: DateType,
                             train_end_date: DateType,
                             holdout_end_date: DateType,
                             split_seed: int,
                             space_holdout_percentage: float,
                             space_column: str,
                             time_column: str,
                             holdout_space: np.ndarray = None,
                             holdout_start_date: DateType = None) -> Tuple[pd.DataFrame, ...]:
    """
    Splits panel data using both ID and Time columns, resulting in four datasets

    1. A training set;
    2. An in training time, but out sample id hold out dataset;
    3. An out of training time, but in sample id hold out dataset;
    4. An out of training time and out of sample id hold out dataset.

    Parameters
    ----------
    dataset : pandas.DataFrame
        A Pandas' DataFrame with an Identifier Column and a Date Column.
        The model will be trained to predict the target column
        from the features.

    train_start_date : str
        A date string representing a the starting time of the training data.
        It should be in the same format as the Date Column in `dataset`.

    train_end_date : str
        A date string representing a the ending time of the training data.
        This will also be used as the start date of the holdout period if no `holdout_start_date` is given.
        It should be in the same format as the Date Column in `dataset`.

    holdout_end_date : str
        A date string representing a the ending time of the holdout data.
        It should be in the same format as the Date Column in `dataset`.

    split_seed : int
        A seed used by the random number generator.

    space_holdout_percentage : float
        The out of id holdout size as a proportion of the in id training
        size.

    space_column : str
        The name of the Identifier column of `dataset`.

    time_column : str
        The name of the Date column of `dataset`.

    holdout_space : np.array
        An array containing the hold out IDs. If not specified,
        A random subset of IDs will be selected for holdout.

    holdout_start_date: str
        A date string representing the starting time of the holdout data.
        If `None` is given it will be equal to `train_end_date`.
        It should be in the same format as the Date Column in `dataset`.

    Returns
    ----------
    train_set : pandas.DataFrame
        The in ID sample and in time training set.

    intime_outspace_hdout : pandas.DataFrame
        The out of ID sample and in time hold out set.

    outime_inspace_hdout : pandas.DataFrame
        The in ID sample and out of time hold out set.

    outime_outspace_hdout : pandas.DataFrame
        The out of ID sample and out of time hold out set.
    """
    holdout_start_date = holdout_start_date if holdout_start_date else train_end_date

    in_time_mask = (dataset[time_column] >= train_start_date) & (dataset[time_column] < train_end_date)
    out_time_mask = (dataset[time_column] >= holdout_start_date) & (dataset[time_column] < holdout_end_date)

    all_space_in_time = dataset[in_time_mask][space_column].unique()

    if holdout_space is None:
        # for repeatability
        state = RandomState(split_seed)
        train_period_space = np.sort(all_space_in_time)

        # randomly sample accounts from the train period to hold out
        partial_holdout_space = state.choice(train_period_space,
                                             int(space_holdout_percentage * len(train_period_space)),
                                             replace=False)

        in_space = pd.Index(all_space_in_time).difference(pd.Index(partial_holdout_space)).values

    else:
        in_space = pd.Index(all_space_in_time).difference(pd.Index(holdout_space)).values

    in_space_mask = dataset[space_column].isin(in_space)

    train_set = dataset[in_space_mask & in_time_mask]
    intime_outspace_hdout = dataset[~in_space_mask & in_time_mask]
    outtime_outspace_hdout = dataset[~in_space_mask & out_time_mask]
    outtime_inspace_hdout = dataset[in_space_mask & out_time_mask]

    return train_set, intime_outspace_hdout, outtime_inspace_hdout, outtime_outspace_hdout


@curry
def stratified_split_dataset(dataset: pd.DataFrame, target_column: str, test_size: float,
                             random_state: Optional[int] = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
        Splits data into a training and testing datasets such that
        they maintain the same class ratio of the original dataset.

        Parameters
        ----------
        dataset : pandas.DataFrame
            A Pandas' DataFrame with the target column.
            The model will be trained to predict the target column
            from the features.

        target_column : str
            The name of the target column of `dataset`.

        test_size : float
            Represent the proportion of the dataset to include in the test split.
            should be between 0.0 and 1.0.

        random_state : int or None, optional (default=None)
            If int, random_state is the seed used by the random number generator;
            If None, the random number generator is the RandomState instance used
            by `np.random`.

        Returns
        ----------
        train_set : pandas.DataFrame
            The train dataset sampled from the full dataset.

        test_set : pandas.DataFrame
            The test dataset sampled from the full dataset.
        """
    train_placeholder = np.zeros(len(dataset))
    target = dataset[target_column]

    splitter = StratifiedShuffleSplit(n_splits=1, test_size=test_size, random_state=random_state)

    train_indices, test_indices = next(splitter.split(train_placeholder, target))
    train_set = dataset.iloc[train_indices]
    test_set = dataset.iloc[test_indices]

    return train_set, test_set



from typing import Any, Dict, List

import pandas as pd
from sklearn.ensemble import IsolationForest
import sklearn
from toolz import curry, merge

from fklearn.common_docstrings import learner_pred_fn_docstring, learner_return_docstring
from fklearn.types import LearnerReturnType
from fklearn.training.utils import log_learner_time, expand_features_encoded


@curry
@log_learner_time(learner_name='isolation_forest_learner')
def isolation_forest_learner(df: pd.DataFrame,
                             features: List[str],
                             params: Dict[str, Any] = None,
                             prediction_column: str = "prediction",
                             encode_extra_cols: bool = True) -> LearnerReturnType:
    """
    Fits an anomaly detection algorithm (Isolation Forest) to the dataset

    Parameters
    ----------
    df : pandas.DataFrame
        A Pandas' DataFrame with features and target columns.
        The model will be trained to predict the target column
        from the features.

    features : list of str
        A list os column names that are used as features for the model. All this names
        should be in `df`.

    params : dict
        The IsolationForest parameters in the format {"par_name": param}. See:
        http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.IsolationForest.html

    prediction_column : str
        The name of the column with the predictions from the model.

    encode_extra_cols : bool (default: True)
        If True, treats all columns in `df` with name pattern fklearn_feat__col==val` as feature columns.
    """

    default_params = {"n_jobs": -1, "random_state": 1729, "contamination": 0.1, "behaviour": "new"}
    params = default_params if not params else merge(default_params, params)

    features = features if not encode_extra_cols else expand_features_encoded(df, features)

    model = IsolationForest()
    model.set_params(**params)
    model.fit(df[features].values)

    def p(new_df: pd.DataFrame) -> pd.DataFrame:
        output_col = {prediction_column: model.decision_function(
            new_df[features])}

        return new_df.assign(**output_col)

    p.__doc__ = learner_pred_fn_docstring("isolation_forest_learner")

    log = {'isolation_forest_learner': {
        'features': features,
        'parameters': params,
        'prediction_column': prediction_column,
        'package': "sklearn",
        'package_version': sklearn.__version__,
        'training_samples': len(df)}}

    return p, p(df), log


isolation_forest_learner.__doc__ += learner_return_docstring("Isolation Forest")



from datetime import datetime
from typing import Any, Callable, Dict, List, Tuple, Union

import pandas as pd

# Date type (used to filter datetime columns in dataframes)
DateType = Union[pd.Period, datetime, str]

# Log types
LogType = Dict[str, Any]
LogListType = List[LogType]
ListLogListType = List[LogListType]

# Learner types
PredictFnType = Callable[[pd.DataFrame], pd.DataFrame]
LearnerLogType = Dict[str, LogType]
LearnerReturnType = Tuple[PredictFnType, pd.DataFrame, LearnerLogType]

UncurriedLearnerFnType = Callable[..., LearnerReturnType]
LearnerFnType = Callable[[pd.DataFrame], LearnerReturnType]

# Evaluator types
EvalReturnType = Dict[str, Union[float, Dict]]
UncurriedEvalFnType = Callable[..., EvalReturnType]
EvalFnType = Callable[[pd.DataFrame], EvalReturnType]

# Splitter types
FoldType = List[Tuple[pd.Index, List[pd.Index]]]
SplitterReturnType = Tuple[FoldType, LogListType]
SplitterFnType = Callable[[pd.DataFrame], SplitterReturnType]

# Validator types
ValidatorReturnType = Dict[str, Union[LogType, LogListType]]

# Perturbator types
PerturbFnType = Callable[[pd.DataFrame], pd.DataFrame]
ColumnWisePerturbFnType = Callable[[pd.Series], pd.Series]

# Extractor types
ExtractorFnType = Callable[[str], float]



from collections import OrderedDict
from itertools import product
from typing import Callable, List

from numpy.random import seed
import pandas as pd
from toolz import curry, partial

from fklearn.validation.validator import parallel_validator, validator
from fklearn.types import EvalFnType, LearnerFnType, LogType, SplitterFnType, ValidatorReturnType

SaveIntermediaryFnType = Callable[[ValidatorReturnType], None]


@curry
def random_search_tuner(space: LogType,
                        train_set: pd.DataFrame,
                        param_train_fn: Callable[[LogType], LearnerFnType],
                        split_fn: SplitterFnType,
                        eval_fn: EvalFnType,
                        iterations: int,
                        random_seed: int = 1,
                        save_intermediary_fn: SaveIntermediaryFnType = None,
                        n_jobs: int = 1) -> List[ValidatorReturnType]:
    """
    Runs several training functions with each run taken from the parameter space

    Parameters
    ----------
    space : dict
        A dictionary with keys as parameter for the model and values as callable that return a parameter.
        Callable must take no parameters and can return always a constant value.
        Example::

            space = {
                'learning_rate': lambda: np.random.choice([1e-3, 1e-2, 1e-1, 1, 10]),
                'num_estimators': lambda: np.random.choice([20, 100, 150])
                }

    train_set : pd.DataFrame
        The training set

    param_train_fn : function(space, train_set) ->  p, new_df, train_log
        A curried training function that os only function of the parameters for the model and the training set.
        Example::

            @curry
            def param_train_fn(space, train_set):
                return xgb_classification_learner(features=["x"],
                                                  target="target",
                                                  learning_rate=space["learning_rate"],
                                                  num_estimators=space["num_estimators"])(train_set)

    split_fn : function(dataset) -> list of folds
        Partially defined split function that takes a dataset and returns
        a list of folds. Each fold is a Tuple of arrays. The fist array in
        each tuple contains training indexes while the second array
        contains validation indexes.
        Examples::

            out_of_time_and_space_splitter(n_splits=n_splits,
                                           in_time_limit=in_time_limit,
                                           space_column=space_column,
                                           time_column=time_column)

    eval_fn : function(dataset) -> eval_log
        A base evaluation function that returns a simple evaluation log. Can't be a spited or the extractor won't work.
        Example: roc_auc_evaluator(target_column="target")

    iterations : int
        The number of iterations to run the parameter tuner

    random_seed : int
        Random seed

    save_intermediary_fn : function(log) -> save to file
        Partially defined saver function that receives a log result from a
        tuning step and appends it into a file
        Example: save_intermediary_result(save_path='tuning.pkl')

    n_jobs : int
        Number of parallel processes to spawn when evaluating a training function

    Returns
    ----------
    tuning_log : list of dict
        A list of tuning log, each containing a training log and a validation log.
    """
    validation_fn = partial(parallel_validator, n_jobs=n_jobs) if n_jobs > 1 else validator

    def tune_iteration() -> ValidatorReturnType:
        iter_space = {k: space[k]() for k in space}
        train_fn = param_train_fn(iter_space)
        validator_log = validation_fn(train_data=train_set, split_fn=split_fn, train_fn=train_fn, eval_fn=eval_fn)

        if save_intermediary_fn is not None:
            save_intermediary_fn(validator_log)

        return validator_log

    seed(random_seed)

    return [tune_iteration() for _ in range(iterations)]


@curry
def grid_search_cv(space: LogType,
                   train_set: pd.DataFrame,
                   param_train_fn: Callable[[LogType], LearnerFnType],
                   split_fn: SplitterFnType,
                   eval_fn: EvalFnType,
                   save_intermediary_fn: SaveIntermediaryFnType = None,
                   load_intermediary_fn: Callable[[str], List[ValidatorReturnType]] = None,
                   warm_start_file: str = None,
                   n_jobs: int = 1) -> List[ValidatorReturnType]:
    """
    Runs several training functions with each run taken from the parameter space

    Parameters
    ----------
    space : dict
        A dictionary with keys as parameter for the model and values as callable that return a parameter.
        Callable must take no parameters and can return always a constant value.
        Example::

            space = {
                'learning_rate': lambda: [1e-3, 1e-2, 1e-1, 1, 10],
                'num_estimators': lambda: [20, 100, 150]
                }

    train_set : pd.DataFrame
        The training set

    param_train_fn : function(space, train_set) ->  p, new_df, train_log
        A curried training function that os only function of the parameters for the model and the training set.
        Example::

            @curry
            def param_train_fn(space, train_set):
                return xgb_classification_learner(features=["x"],
                                                  target="target",
                                                  learning_rate=space["learning_rate"],
                                                  num_estimators=space["num_estimators"])(train_set)

    split_fn : function(dataset) -> list of folds
        Partially defined split function that takes a dataset and returns
        a list of folds. Each fold is a Tuple of arrays. The fist array in
        each tuple contains training indexes while the second array
        contains validation indexes.
        Examples::

            out_of_time_and_space_splitter(n_splits=n_splits,
                                           in_time_limit=in_time_limit,
                                           space_column=space_column,
                                           time_column=time_column)

    eval_fn : function(dataset) -> eval_log
        A base evaluation function that returns a simple evaluation log. Can't be a spited or the extractor won't work.
        Example: roc_auc_evaluator(target_column="target")

    save_intermediary_fn : function(log) -> save to file
        Partially defined saver function that receives a log result from a
        tuning step and saves it into a file
        Example: save_intermediary_result(save_path='tuning.pkl')

    load_intermediary_fn : function(path) -> save to file
        Partially defined load function that receives a path and loads previous logs
        from this file
        Example: load_intermediary_result('tuning.pkl')

    warm_start_file: str
        File containing intermediary results for grid search. If this file
        is present, we will perform grid search from the last combination of
        parameters.

    n_jobs : int
        Number of parallel processes to spawn when evaluating a training function


    Returns
    ----------
    tuning_log : list of dict
        A list of tuning log, each containing a training log and a validation log.
    """

    validation_fn = partial(parallel_validator, n_jobs=n_jobs) if n_jobs > 1 else validator

    def tune_iteration(iter_space: LogType) -> ValidatorReturnType:
        train_fn = param_train_fn(iter_space)
        validator_log = validation_fn(train_data=train_set, split_fn=split_fn, train_fn=train_fn, eval_fn=eval_fn)
        validator_log['iter_space'] = OrderedDict(sorted(iter_space.items()))

        if save_intermediary_fn is not None:
            save_intermediary_fn(validator_log)

        return validator_log

    sorted_space_keys = sorted(space.keys())
    params = (space[k]() for k in sorted_space_keys)
    combinations = set(product(*params))

    if warm_start_file is not None and load_intermediary_fn is not None:
        results = load_intermediary_fn(warm_start_file)
        computed_combs = set([tuple(log['iter_space'].values()) for log in results])  # type: ignore
        combinations = combinations.difference(computed_combs)

    return [tune_iteration({k_v[0]: k_v[1] for k_v in zip(sorted_space_keys, comb)}) for comb in combinations]
